<?php
/**
 * Define the WC_AvaTax_API_Tax_Response class
 *
 * This source file is subject to the GNU General Public License v3.0
 * that is bundled with this package in the file license.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.gnu.org/licenses/gpl-3.0.html
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@skyverge.com so we can send you a copy immediately.
 *
 * DISCLAIMER
 *
 * Do not edit or add to this file if you wish to upgrade WooCommerce AvaTax to newer
 * versions in the future. If you wish to customize WooCommerce AvaTax for your
 * needs please refer to http://docs.woocommerce.com/document/woocommerce-avatax/
 *
 * @package   AvaTax\API
 * @author    SkyVerge
 * @copyright Copyright (c) 2016-2018, SkyVerge, Inc.
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU General Public License v3.0
 */

defined( 'ABSPATH' ) or exit;

/**
 * The AvaTax API tax response class.
 *
 * @since 1.0.0
 */
class WC_AvaTax_API_Tax_Response extends WC_AvaTax_API_Response {


	/**
	 * Gets any cart item lines.
	 *
	 * This only returns products and no fees or shipping.
	 *
	 * @since 1.5.1
	 *
	 * @return array
	 */
	public function get_cart_lines() {

		$lines = $this->get_lines();

		foreach ( $lines as $key => $line ) {

			if ( SV_WC_Helper::str_starts_with( $line['id'], 'fee_' ) || SV_WC_Helper::str_starts_with( $line['id'], 'shipping_' ) ) {
				unset( $lines[ $key ] );
			}
		}

		return $lines;
	}


	/**
	 * Gets any fee lines.
	 *
	 * @since 1.5.0
	 *
	 * @return array
	 */
	public function get_fee_lines() {

		$lines = $this->get_lines();

		foreach ( $lines as $key => $line ) {

			if ( ! SV_WC_Helper::str_starts_with( $line['id'], 'fee_' ) && ! in_array( $line['id'], array( 'ImportDuties', 'ImportFees' ) ) ) {
				unset( $lines[ $key ] );
			} else {
				$lines[ $key ]['id'] = str_replace( 'fee_', '', $line['id'] );
			}
		}

		return $lines;
	}


	/**
	 * Gets any shipping lines.
	 *
	 * @since 1.5.0
	 *
	 * @return array
	 */
	public function get_shipping_lines() {

		$lines = $this->get_lines();

		foreach ( $lines as $key => $line ) {

			if ( SV_WC_Helper::str_starts_with( $line['id'], 'shipping_' ) ) {
				$lines[ $key ]['id'] = str_replace( 'shipping_', '', $line['id'] );
			} else {
				unset( $lines[ $key ] );
			}
		}

		return $lines;
	}


	/**
	 * Get the calculated line items.
	 *
	 * @since 1.0.0
	 * @return array The calculated line items
	 */
	public function get_lines() {

		$lines = array();

		foreach ( $this->lines as $line ) {

			$lines[] = array(
				'id'          => $line->lineNumber,
				'sku'         => isset( $line->itemCode ) ? $line->itemCode : '',
				'amount'      => $line->lineAmount,
				'tax'         => $line->tax,
				'code'        => $line->taxCode,
				'rates'       => $this->build_rates( $line->details ),
				'origin'      => isset( $line->originAddressId ) ? $this->get_origin_address( $line ) : array(),
				'destination' => isset( $line->destinationAddressId ) ? $this->get_destination_address( $line ) : array(),
			);
		}

		return $lines;
	}


	/**
	 * Get the origin address.
	 *
	 * @since 1.0.0
	 * @return array The origin address
	 */
	public function get_origin_address( $line = null ) {

		$item           = ( $line ) ? $line : $this;
		$origin_address = array();

		// Get the origin address
		foreach ( $this->get_addresses() as $address ) {

			if ( $item->originAddressId === $address->id ) {

				// Map the API response to their proper keys
				$origin_address = array(
					'address_1' => $address->line1,
					'address_2' => $address->line2,
					'city'      => $address->city,
					'state'     => $address->region,
					'country'   => $address->country,
					'postcode'  => $address->postalCode,
				);

				break;
			}
		}

		return $origin_address;
	}


	/**
	 * Get the destination address.
	 *
	 * @since 1.1.0
	 * @return array The destination address
	 */
	public function get_destination_address( $line = null ) {

		$item                = ( $line ) ? $line : $this;
		$destination_address = array();

		// Get the destination address
		foreach ( $this->get_addresses() as $address ) {

			if ( $item->destinationAddressId === $address->id ) {

				// Map the API response to their proper keys
				$destination_address = array(
					'address_1' => $address->line1,
					'address_2' => $address->line2,
					'city'      => $address->city,
					'state'     => $address->region,
					'country'   => $address->country,
					'postcode'  => $address->postalCode,
				);

				break;
			}
		}

		return $destination_address;
	}


	/**
	 * Get the transaction addresses.
	 *
	 * @since 1.5.0
	 *
	 * @return array
	 */
	public function get_addresses() {

		return $this->addresses;
	}


	/**
	 * Get the total tax amount.
	 *
	 * @since 1.0.0
	 * @return float The total tax calculated.
	 */
	public function get_total_tax() {

		return $this->totalTax;
	}


	/**
	 * Determines if the line amounts included tax.
	 *
	 * @since 1.5.0
	 *
	 * @return bool
	 */
	public function is_tax_included() {

		$tax_included = false;

		foreach ( $this->lines as $line ) {

			// if one is, they all are
			if ( $line->taxIncluded ) {
				$tax_included = true;
				break;
			}
		}

		return $tax_included;
	}


	/**
	 * Get the effective tax date.
	 *
	 * @since 1.0.0
	 * @return string The effective tax date in YYYY-MM-DD format.
	 */
	public function get_tax_date() {

		return date( 'Y-m-d', strtotime( $this->taxDate ) );
	}


	/**
	 * Gets all rates use to calculate tax for the transaction.
	 *
	 * @since 1.5.0
	 *
	 * @return array $rates \WC_AvaTax_API_Tax_Rate rate objects
	 */
	public function get_rates() {

		return $this->build_rates( $this->summary );
	}


	/**
	 * Builds the array of rate objects based on the response data.
	 *
	 * @since 1.5.0
	 *
	 * @param array $raw_rates rate data from the AvaTax API
	 *
	 * @return array $rates \WC_AvaTax_API_Tax_Rate rate objects
	 */
	protected function build_rates( $raw_rates ) {

		$rates = array();

		if ( is_array( $raw_rates ) ) {

			$count = 1;

			foreach ( $raw_rates as $rate ) {

				$name = 'LandedCost' === $rate->taxType ? __( 'Customs Import & Duty', 'woocommerce-avatax' ) : trim( "{$rate->jurisType} {$rate->taxType}" );

				$rate = new WC_AvaTax_API_Tax_Rate( array(
					'code'  => "{$count}-{$rate->taxName}",
					'name'  => $name,
					'rate'  => $rate->rate,
					'total' => $rate->taxCalculated,
				) );

				$rates[ $rate->get_code() ] = $rate;

				$count++;
			}
		}

		return $rates;
	}


	/**
	 * Gets the calculation response messages.
	 *
	 * These can be general tax calculation result messages (which we don't
	 * always need) or Landed Cost notes, like the duty to be paid when the
	 * customer is importer of record.
	 *
	 * @since 1.5.0
	 *
	 * @return array
	 */
	public function get_messages() {

		$messages = $this->messages;

		if ( ! is_array( $messages ) ) {
			$messages = array();
		}

		return $messages;
	}


}
